import os

import rclpy
from rclpy.node import Node

from transitions import Machine
import cma
import yaml
import numpy as np
from src.interaction_utils.serialization import flatten_population, unflatten_population

# Msg/Srv/Action
from interaction_msgs.srv import Query
from interaction_msgs.srv import TaskEvaluation
from std_msgs.msg import Bool


class CMAESOptimizationNode(Node):
    def __init__(self):
        super().__init__('cmaes_optimization_node')

        # CMA-ES Attributes
        self.cmaes = None
        self.number_of_dimensions = 3
        self.number_of_parameters_per_dimensions = 10
        # the number of weights is double the number of dims * params per dim since its Position and Velocity
        self.number_of_weights = 2 * self.number_of_dimensions * self.number_of_parameters_per_dimensions

        self.random_seed = None

        # Query Attributes
        self.query_metric = 'random'
        self.query_parameters = {}
        self.best_rewards_per_iteration = []
        # ROS2 Interfaces
        self.__future = None
        self.__task_response = None
        # Heartbeat Topics - to make sure that there is no deadlock
        self.heartbeat_timeout = 30     # secs
        self.heartbeat_timer = self.create_timer(1.0, self.check_heartbeats)

        # MR Heartbeat
        self.last_mr_heartbeat_time = None
        self.mr_heartbeat_sub = self.create_subscription(Bool, 'interaction/mr_heartbeat', self.mr_heatbeat_callback)
        self.last_task_heartbeat_time = None
        self.task_heartbeat_sub = self.create_subscription(Bool, 'interaction/task_heartbeat', self.task_heartbeat_callback)

        # Topics

        # Service
        self.query_srv = self.create_client(Query, 'query_srv')
        self.task_srv = self.create_client(TaskEvaluation, 'task_srv')

        # Define states
        self.states = [
            'idle',
            'initialization',
            'query_decision',
            'non_interactive_mode',
            'wait_for_evaluation',
            'interactive_mode',
            'wait_for_user_response',
            'prepare_user_data_for_evaluation',
            'wait_for_user_evaluation',
            'update_optimizer',
            'check_termination',
            'complete',
            'error_recovery',
        ]

        # Initialize state machine
        self.machine = Machine(self, states=self.states, initial='idle')

        # region Transitions
        self.machine.add_transition(trigger='order_received', source='idle', dest='initialization')
        self.machine.add_transition(trigger='initialization_complete', source='initialization', dest='query_decision')
        self.machine.add_transition(trigger='no_interaction', source='query_decision', dest='non_interactive_mode')
        self.machine.add_transition(trigger='data_send_for_evaluation', source='non_interactive_mode', dest='wait_for_evaluation')
        self.machine.add_transition(trigger='evaluation_response_received', source='wait_for_evaluation', dest='update_optimizer')
        self.machine.add_transition(trigger='interaction', source='query_decision', dest='interactive_mode')
        self.machine.add_transition(trigger='data_send_to_user', source='interactive_mode', dest='wait_for_user_response')
        self.machine.add_transition(trigger='user_response_received', source='wait_for_user_response', dest='prepare_user_data_for_evaluation')
        self.machine.add_transition(trigger='send_user_data_to_evaluation', source='prepare_user_data_for_evaluation', dest='wait_for_user_evaluation')
        self.machine.add_transition(trigger='received_user_data_for_evaluation', source='wait_for_user_evaluation', dest='update_optimizer')
        self.machine.add_transition(trigger='optimizer_updated', source='update_optimizer', dest='check_termination')
        self.machine.add_transition(trigger='next_optimizer_step', source='check_termination', dest='query_decision')
        self.machine.add_transition(trigger='finished', source='check_termination', dest='complete')
        self.machine.add_transition(trigger='error_trigger', source='*', dest='error_recovery')
        self.machine.add_transition(trigger='recovery_complete', source='error_recovery', dest='idle')
        # endregion

    # region State Functions
    def on_enter_initialization(self):
        config_file_path = self.get_parameter('cmaes_config_file_path').get_parameter_value().string_value

        # Load YAML
        with open(config_file_path, 'r') as file:
            config = yaml.safe_load(file)

        if config['random_seed'] == '':
            self.random_seed = None
        else:
            self.random_seed = int(config['random_seed'])
            config['seed'] = self.random_seed

        mean_centre = config['initial_mean_centre']
        mean_std_dev = config['initial_mean_std_dev']

        random_gen = np.random.default_rng(seed=self.random_seed)

        initial_mean = random_gen.normal(mean_centre, mean_std_dev, size=(self.number_of_weights, 1))
        initial_variance = config['initial_variance']
        self.cmaes = cma.CMAEvolutionStrategy(initial_mean, initial_variance, inopts=config)

        # Trigger transition
        self.initialization_complete()

    def on_enter_query_decision(self):
        request = Query.Request()

        if self.query_metric == 'random':
            pass
        elif self.query_metric == 'regular':
            pass
        elif self.query_metric == 'improvement':
            pass

        self.__future = self.query_srv.call_async(request)
        self.__future.add_done_callback(self.handle_query_response)

    def on_enter_non_interactive_mode(self):
        # Reset Task heartbeat to check if the other node crashed
        self.last_task_heartbeat_time = None

        request = TaskEvaluation.Request()
        request.user_input = False
        request.number_of_population = self.cmaes.popsize
        request.number_of_dimensions = self.number_of_dimensions * 2
        request.number_of_parameters_per_dimensions = self.number_of_parameters_per_dimensions

        population = self.cmaes.ask()
        flat_population = flatten_population(population)

        request.parameter_array = flat_population

        self.__future = self.task_srv.call_async(request)
        self.__future.add_done_callback(self.handle_task_response)

        self.data_send_for_evaluation()

    def on_enter_interactive_mode(self):
        pass

    def on_enter_prepare_user_data_for_evaluation(self):
        pass

    def on_enter_update_optimizer(self):
        pass

    def on_enter_check_termination(self):
        pass

    def on_enter_error_recovery(self):
        pass

    # endregion

    # region Callback Functions
    def check_heartbeats(self):
        if self.last_mr_heartbeat_time:
            current_time = self.get_clock().now().nanoseconds
            if (current_time - self.last_mr_heartbeat_time) > (self.heartbeat_timeout * 1e9):
                self.get_logger().error("MR Interface heartbeat timed out!")
                self.error_trigger()
        else:
            pass

        if self.last_mr_heartbeat_time:
            current_time = self.get_clock().now().nanoseconds
            if (current_time - self.last_task_heartbeat_time) > (self.heartbeat_timeout * 1e9):
                self.get_logger().error("Task Node heartbeat timed out!")
                self.error_trigger()
        else:
            pass

    def mr_heartbeat_callback(self, _):
        self.last_mr_heartbeat_time = self.get_clock().now().nanoseconds

    def handle_query_response(self, future):
        try:
            response = future.result()

            if response.interaction:
                self.interaction()
            else:
                self.non_interaction()

        except Exception as e:
            self.get_logger().error(f'Query service call failed: {e}')
            self.error_trigger()

    def handle_task_response(self, future):
        try:
            self.__task_response = future.result()
            self.evaluation_response_received()

        except Exception as e:
            self.get_logger().error(f'Task service call failed: {e}')
            self.error_trigger()

    # endregion